# Copyright (c) 2019 Jonathan de Bruin

#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation
#  files (the "Software"), to deal in the Software without
#  restriction, including without limitation the rights to use,
#  copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following
#  conditions:

#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.

#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

"""Statistics Netherlands opendata CLI client"""

import argparse
import json
import sys

import cbsodata

AVAILABLE_CMDS = ["data", "info", "list"]


def json_outputter(data_obj, max_rows=None):
    """Print data in JSON format."""

    # cut of the at max_rows
    if isinstance(max_rows, (int, float)):
        data_obj = data_obj[0:max_rows]

    for line in data_obj:
        print(json.dumps(line))


def text_outputter(data_obj, max_rows=None):
    """Print data in text format."""

    # cut of the at max_rows
    if isinstance(max_rows, (int, float)):
        data_obj = data_obj[0:max_rows]

    # collect the maximum length in each column
    value_max_len = {}
    for d in data_obj:
        for key, value in d.items():
            try:
                if len(str(value)) > value_max_len[key]:
                    value_max_len[key] = len(str(value))
            except KeyError:
                value_max_len[key] = len(str(value))

    # get a list of columns
    columns = value_max_len.keys()

    # check if the column name is larger than the largest value
    for col in columns:
        if len(str(col)) > value_max_len[col]:
            value_max_len[col] = len(str(col))

    # print the column line
    col_line = ""
    for i, col in enumerate(columns):
        if i != (len(columns) - 1):
            col_line = col_line + str(col) \
                .upper().ljust(value_max_len[col] + 2)
        else:
            col_line = col_line + str(col).upper()
    print(col_line)

    # print the data
    for d in data_obj:
        line = ""
        for i, col in enumerate(columns):
            if i != (len(columns) - 1):
                try:
                    line = line + str(d[col]).ljust(value_max_len[col] + 2)
                except KeyError:
                    line = " " * (value_max_len[col] + 2)
            else:
                try:
                    line = line + str(d[col])
                except KeyError:
                    pass
        print(line)


def parse_argument_table_id(parser):
    parser.add_argument(
        "table_id",
        type=str,
        help="table identifier"
    )


def parse_argument_catalog(parser):
    parser.add_argument(
        "--catalog_url",
        default=None,
        type=str,
        help="the catalog to download the data from")


def parse_argument_output_format(parser):
    parser.add_argument(
        "--output_format", "-f",
        default="json",
        type=str,
        help="format to show table ('json', 'text')")


def parse_argument_output(parser):
    parser.add_argument(
        "--output_file", "-o",
        default=None,
        type=str,
        help="file to store the output (only json support)")


def parse_argument_max_rows(parser):
    parser.add_argument(
        "--max_rows", "-n",
        default=100,
        type=int,
        help="maximum number of rows to output")


def save_list_to_json(data_obj, fp):
    """Write list with dicts to json"""

    with open(fp, 'w+') as f:
        for line in data_obj:
            f.write(json.dumps(line) + "\n")


def main():

    if len(sys.argv) > 1 and sys.argv[1] == "data":
        parser = argparse.ArgumentParser(
            prog="cbsodata",
            description="""
                CBS Open Data: Command Line Interface

                Get data by table identifier.
            """
        )
        parse_argument_table_id(parser)
        parse_argument_catalog(parser)
        parse_argument_output_format(parser)
        parse_argument_max_rows(parser)
        parse_argument_output(parser)
        args = parser.parse_args(sys.argv[2:])

        result = cbsodata.get_data(args.table_id, catalog_url=args.catalog_url)

        if args.output_file:
            save_list_to_json(result, args.output_file)

        if args.output_format == "text":
            text_outputter(result, max_rows=args.max_rows)
        else:
            json_outputter(result, max_rows=args.max_rows)

    elif len(sys.argv) > 1 and sys.argv[1] == "info":
        parser = argparse.ArgumentParser(
            prog="cbsodata",
            description="""
                CBS Open Data: Command Line Interface

                Get data infomation by table identifier.
            """
        )
        parse_argument_table_id(parser)
        parse_argument_catalog(parser)
        parse_argument_output_format(parser)
        parse_argument_output(parser)
        args = parser.parse_args(sys.argv[2:])

        result = cbsodata.get_info(args.table_id, catalog_url=args.catalog_url)

        if args.output_file:
            with open(args.output_file, 'w') as f:
                json.dump(result, f, indent=4)

        if args.output_format == "text":
            text_outputter(
                [{"Label": k, "Value": v} for k, v in result.items()]
            )
        else:
            print(json.dumps(result, indent=4))

    elif len(sys.argv) > 1 and sys.argv[1] == "list":
        parser = argparse.ArgumentParser(
            prog="cbsodata",
            description="""
                CBS Open Data: Command Line Interface

                Get list of available tables.
            """
        )
        parse_argument_catalog(parser)
        parse_argument_output_format(parser)
        parse_argument_max_rows(parser)
        parse_argument_output(parser)
        args = parser.parse_args(sys.argv[2:])

        result = cbsodata.get_table_list(catalog_url=args.catalog_url)

        if args.output_file:
            save_list_to_json(result, args.output_file)

        if args.output_format == "text":
            text_outputter(result, max_rows=args.max_rows)
        else:
            json_outputter(result, max_rows=args.max_rows)

    # no valid sub command
    else:
        parser = argparse.ArgumentParser(
            prog="cbsodata",
            description="""
                CBS Open Data: Command Line Interface
            """
        )
        parser.add_argument(
            "subcommand",
            nargs="?",
            type=lambda x: isinstance(x, str) and x in AVAILABLE_CMDS,
            help="the subcommand (one of '{}')".format(
                "', '".join(AVAILABLE_CMDS))
        )
        parser.add_argument(
            "--version",
            action='store_true',
            help="show the package version")

        args = parser.parse_args()

        if args.subcommand is None:
            parser.print_help()


if __name__ == '__main__':
    main()



# Copyright (c) 2019 Jonathan de Bruin

#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation
#  files (the "Software"), to deal in the Software without
#  restriction, including without limitation the rights to use,
#  copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following
#  conditions:

#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.

#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

"""Statistics Netherlands opendata API client for Python"""

from cbsodata.cbsodata3 import *

__version__ = "1.2.2"



# Copyright (c) 2016 Jonathan de Bruin

#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation
#  files (the "Software"), to deal in the Software without
#  restriction, including without limitation the rights to use,
#  copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following
#  conditions:

#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.

#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

"""Statistics Netherlands opendata API client for Python"""

__all__ = ['download_data', 'get_data', 'get_info', 'get_meta',
           'get_table_list', 'options', 'catalog']

import os
import json
import copy
import logging
from contextlib import contextmanager

import requests
from requests import Session, Request


CBSOPENDATA = "opendata.cbs.nl"  # deprecate in next version
API = "ODataApi/odata"
BULK = "ODataFeed/odata"

CATALOG = "ODataCatalog"
FORMAT = "json"


class OptionsManager(object):
    """Class for option management"""

    def __init__(self):

        self.use_https = True
        self.api_version = "3"

        # Enable in next version
        # self.catalog_url = "opendata.cbs.nl"

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "catalog_url = {}, use_https = {}".format(
            self.catalog_url, self.use_https)

    def __getitem__(self, arg):
        return getattr(self, arg)

    def __setitem__(self, arg, value):
        setattr(self, arg, value)

    def _log_setting_change(self, setting_name, old_value, new_value):
        logging.info(
            "Setting '{}' changed from '{}' to '{}'.".format(
                setting_name, old_value, new_value)
        )

    def __getattr__(self, arg):
        return getattr(self, arg)

    def __setattr__(self, arg, value):
        try:
            old_value = copy.copy(getattr(self, arg))
        except Exception:
            old_value = "undefined"

        self._log_setting_change(arg, old_value, value)
        super(OptionsManager, self).__setattr__(arg, value)

    @property
    def catalog_url(self):
        return CBSOPENDATA

    @catalog_url.setter
    def catalog_url(self, url):
        global CBSOPENDATA
        CBSOPENDATA = url  # noqa


# User options
options = OptionsManager()


def _get_catalog_url(url):

    return options.catalog_url if url is None else url


def _get_table_url(table_id, catalog_url=None):
    """Create a table url for the given table identifier."""

    if catalog_url is None:
        _catalog_url = options.catalog_url
    else:
        _catalog_url = catalog_url

    components = {"http": "https://" if options.use_https else "http://",
                  "baseurl": _catalog_url,
                  "bulk": BULK,
                  "table_id": table_id}

    # http://opendata.cbs.nl/ODataApi/OData/37506wwm
    return "{http}{baseurl}/{bulk}/{table_id}/".format(**components)


def _download_metadata(table_id, metadata_name, select=None, filters=None,
                       catalog_url=None):
    """Download metadata."""

    # http://opendata.cbs.nl/ODataApi/OData/37506wwm/UntypedDataSet?$format=json
    url = _get_table_url(table_id, catalog_url=catalog_url) + metadata_name

    params = {}
    params["$format"] = FORMAT

    if select:
        params['$select'] = _select(select)
    if filters:
        params['$filter'] = _filters(filters)

    try:
        data = []

        while (url is not None):

            s = Session()
            p = Request('GET', url, params=params).prepare()

            logging.info("Download " + p.url)

            r = s.send(p)
            r.raise_for_status()

            res = r.json(encoding='utf-8')
            data.extend(res['value'])

            try:
                url = res['odata.nextLink']
            except KeyError:
                url = None

        return data

    except requests.HTTPError as http_err:
        raise requests.HTTPError(
            "Downloading table '{}' failed. {}".format(table_id, str(http_err))
        )


def _save_data(data, dir, metadata_name):
    """Save the data."""

    if not os.path.exists(dir):
        os.makedirs(dir)

    fp = os.path.join(dir, metadata_name + '.json')

    with open(fp, 'w') as output_file:
        json.dump(data, output_file, indent=2)


def _filters(query):
    """Filter rows with a CBS-style query.

    Parameters
    ----------
    query : str
        The rows to return.

    Returns
    -------
    str
        Filter parameter for URL
    """

    return query


def _select(select):
    """Select columns.

    Parameters
    ----------
    select : str
        The columns to return.

    Returns
    -------
    str
        Select parameter for URL
    """

    if isinstance(select, list):
        select = ','.join(select)

    return select


def download_data(table_id, dir=None, typed=False, select=None, filters=None,
                  catalog_url=None):
    """Download the CBS data and metadata.

    Parameters
    ----------
    table_id : str
        The identifier of the table.
    dir : str
        Folder to save data to. If not given, data is not stored on
        disk.
    typed : bool
        Return a typed data table. Default False.
    select : str, list
        Column label or list of column labels to return.
    filters : str
        Return only rows that agree on the filter.
    catalog_url : str
        The url of the catalog. Default "opendata.cbs.nl".

    Returns
    -------
    list
        A dictionary with the (meta)data of the table
    """

    _catalog_url = _get_catalog_url(catalog_url)

    # http://opendata.cbs.nl/ODataApi/OData/37506wwm?$format=json
    metadata_tables = _download_metadata(
        table_id, "", catalog_url=_catalog_url
    )

    # The names of the tables with metadata
    metadata_table_names = [table['name'] for table in metadata_tables]

    # Download only the typed or untyped data
    typed_or_not_str = "TypedDataSet" if typed else "UntypedDataSet"
    metadata_table_names.remove(typed_or_not_str)

    data = {}

    for table_name in metadata_table_names:

        # download table
        if table_name in ["TypedDataSet", "UntypedDataSet"]:
            metadata = _download_metadata(table_id, table_name,
                                          select=select, filters=filters,
                                          catalog_url=_catalog_url)
        else:
            metadata = _download_metadata(table_id, table_name,
                                          catalog_url=_catalog_url)

        data[table_name] = metadata

        # save the data
        if dir:
            _save_data(metadata, dir, table_name)

    return data


def get_table_list(select=None, filters=None, catalog_url=None):
    """Get a list with the available tables.

    Parameters
    ----------
    select : list, str
        Column label or list of column labels to return.
    filters : str
        Return only rows that agree on the filter.
    catalog_url : str
        The url of the catalog. Default "opendata.cbs.nl".

    Returns
    -------
    list
        A list with the description of each table in the catalog.
    """

    # http://opendata.cbs.nl/ODataCatalog/Tables?$format=json&$filter=ShortTit
    # le%20eq%20%27Zeggenschap%20bedrijven;%20banen,%20grootte%27

    # http://opendata.cbs.nl/ODataCatalog/Tables?$format=json

    _catalog_url = _get_catalog_url(catalog_url)

    components = {"http": "https://" if options.use_https else "http://",
                  "baseurl": _catalog_url,
                  "catalog": CATALOG}

    url = "{http}{baseurl}/{catalog}/Tables?$format=json".format(**components)

    params = {}
    if select:
        params['$select'] = _select(select)
    if filters:
        params['$filter'] = _filters(filters)

    try:
        s = Session()
        p = Request('GET', url, params=params).prepare()

        logging.info("Download " + p.url)

        r = s.send(p)
        r.raise_for_status()
        res = r.json()

        return res['value']

    except requests.HTTPError as http_err:
        raise requests.HTTPError(
            "Downloading table list failed. {}".format(str(http_err))
        )


def get_info(table_id, catalog_url=None):
    """Get information about a table.

    Parameters
    ----------
    table_id : str
        The identifier of the table.
    catalog_url : str
        The url of the catalog. Default "opendata.cbs.nl".

    Returns
    -------
    dict
        Table information
    """

    info_list = _download_metadata(
        table_id,
        "TableInfos",
        catalog_url=_get_catalog_url(catalog_url)
    )

    if len(info_list) > 0:
        return info_list[0]
    else:
        return None


def get_meta(table_id, name, catalog_url=None):
    """Get the metadata of a table.

    Parameters
    ----------
    table_id : str
        The identifier of the table.
    name : str
        The name of the metadata (for example DataProperties).
    catalog_url : str
        The url of the catalog. Default "opendata.cbs.nl".

    Returns
    -------
    list
        A list with metadata (dict type)
    """

    return _download_metadata(
        table_id, name, catalog_url=_get_catalog_url(catalog_url)
    )


def get_data(table_id, dir=None, typed=False, select=None, filters=None,
             catalog_url=None):
    """Get the CBS data table.

    Parameters
    ----------
    table_id : str
        The identifier of the table.
    dir : str
        Folder to save data to. If not given, data is not stored
            on disk.
    typed : bool
        Return a typed data table. Default False.
    select : list
        Column label or list of column labels to return.
    filters : str
        Return only rows that agree on the filter.
    catalog_url : str
        The url of the catalog. Default "opendata.cbs.nl".

    Returns
    -------
    list
        The requested data.
    """

    _catalog_url = _get_catalog_url(catalog_url)

    metadata = download_data(table_id, dir=dir, typed=typed,
                             select=select, filters=filters,
                             catalog_url=_catalog_url)

    if "TypedDataSet" in metadata.keys():
        data = metadata["TypedDataSet"]
    else:
        data = metadata["UntypedDataSet"]

    exclude = [
        "TableInfos", "TypedDataSet", "UntypedDataSet",
        "DataProperties", "CategoryGroups"
    ]

    norm_cols = list(set(metadata.keys()) - set(exclude))

    for norm_col in norm_cols:
        metadata[norm_col] = {r['Key']: r for r in metadata[norm_col]}

    for i in range(0, len(data)):

        for norm_col in norm_cols:

            try:
                v = data[i][norm_col]
                data[i][norm_col] = metadata[norm_col][v]['Title']
            except KeyError:
                pass

    return data


@contextmanager
def catalog(catalog_url, use_https=True):
    """Context manager for catalogs.

    Parameters
    ----------
    catalog_url : str
        Url for the catalog. For example:
        dataderden.cbs.nl.
    use_https : bool
        Use https. Default True.

    """

    old_url = copy.copy(options.catalog_url)
    options.catalog_url = catalog_url

    yield

    options.catalog_url = old_url



from setuptools import setup

from os import path
from codecs import open

here = path.abspath(path.dirname(__file__))

# Use readme as long description
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Extract version from cbsodata.py
for line in open(path.join('cbsodata', '__init__.py')):
    if line.startswith('__version__'):
        exec(line)
        break

setup(
    name='cbsodata',
    version=__version__,  # noqa
    description='Statistics Netherlands opendata API client for Python',
    long_description=long_description,
    url='https://github.com/J535D165/cbsodata',
    author='Jonathan de Bruin',
    author_email='jonathandebruinos@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    keywords='cbs statistics odata netherlands dutch',
    install_requires=['requests'],
    entry_points={
        # two entry points: cbs and cbsodata
        'console_scripts': [
            'cbsodata=cbsodata.__main__:main',
            'cbs=cbsodata.__main__:main',
        ],

    },
    py_modules=['cbsodata.cbsodata3', 'cbsodata.__main__'],
    tests_require=[
        'pytest'
    ],
)
